@@ -16,6 +16,7 @@ use ostree::{gio, glib};
 use std::collections::HashMap;
 use std::iter::FromIterator;
 use std::sync::{Arc, Mutex};
+use tokio::sync::mpsc::{Receiver, Sender};
 
 /// Configuration for the proxy.
 ///
@@ -54,6 +55,19 @@ fn ref_for_image(l: &ImageReference) -> Result<String> {
     refescape::prefix_escape_for_ref(IMAGE_PREFIX, &l.to_string())
 }
 
+/// Sent across a channel to track start and end of a container fetch.
+#[derive(Debug)]
+pub enum ImportProgress {
+    /// Started fetching this layer.
+    OstreeChunkStarted(Descriptor),
+    /// Successfully completed the fetch of this layer.
+    OstreeChunkCompleted(Descriptor),
+    /// Started fetching this layer.
+    DerivedLayerStarted(Descriptor),
+    /// Successfully completed the fetch of this layer.
+    DerivedLayerCompleted(Descriptor),
+}
+
 /// State of an already pulled layered image.
 #[derive(Debug, PartialEq, Eq)]
 pub struct LayeredImageState {
@@ -95,6 +109,8 @@ pub struct ImageImporter {
     imgref: OstreeImageReference,
     target_imgref: Option<OstreeImageReference>,
     pub(crate) proxy_img: OpenedImage,
+
+    layer_progress: Option<Sender<ImportProgress>>,
 }
 
 /// Result of invoking [`LayeredImageImporter::prepare`].
@@ -274,6 +290,7 @@ impl ImageImporter {
             proxy_img,
             target_imgref: None,
             imgref: imgref.clone(),
+            layer_progress: None,
         })
     }
 
@@ -286,6 +303,14 @@ impl ImageImporter {
         self.prepare_internal(false).await
     }
 
+    /// Create a channel receiver that will get notifications for layer fetches.
+    pub fn request_progress(&mut self) -> Receiver<ImportProgress> {
+        assert!(self.layer_progress.is_none());
+        let (s, r) = tokio::sync::mpsc::channel(2);
+        self.layer_progress = Some(s);
+        r
+    }
+
     /// Determine if there is a new manifest, and if so return its digest.
     #[context("Fetching manifest")]
     pub(crate) async fn prepare_internal(&mut self, verify_layers: bool) -> Result<PrepareResult> {
@@ -329,22 +354,14 @@ impl ImageImporter {
 
         let label = crate::container::OSTREE_DIFFID_LABEL;
         let config_labels = config.config().as_ref().and_then(|c| c.labels().as_ref());
+        let diffid = config_labels.and_then(|labels| labels.get(label));
         // For backwards compatibility, if there's only 1 layer, don't require the label.
         // This can be dropped when we drop format version 0 support.
-        let commit_layer_digest = if config.rootfs().diff_ids().len() == 1 {
-            manifest.layers()[0].digest()
-        } else {
-            let diffid = config_labels
-                .and_then(|labels| labels.get(label))
-                .ok_or_else(|| {
-                    anyhow!(
-                        "Missing annotation {} (not an ostree-exported container?)",
-                        label
-                    )
-                })?;
-
+        let commit_layer_digest = if let Some(diffid) = diffid {
             let layer = layer_from_diffid(&manifest, &config, diffid.as_str())?;
             layer.digest()
+        } else {
+            manifest.layers()[0].digest()
         };
         let mut component_layers = Vec::new();
         let mut commit_layer = None;
@@ -405,6 +422,10 @@ impl ImageImporter {
             if layer.commit.is_some() {
                 continue;
             }
+            if let Some(p) = self.layer_progress.as_ref() {
+                p.send(ImportProgress::OstreeChunkStarted(layer.layer.clone()))
+                    .await?;
+            }
             let (blob, driver) =
                 fetch_layer_decompress(&mut self.proxy, &self.proxy_img, &layer.layer).await?;
             let blob = super::unencapsulate::ProgressReader {
@@ -433,8 +454,18 @@ impl ImageImporter {
                 });
             let commit = super::unencapsulate::join_fetch(import_task, driver).await?;
             layer.commit = commit;
+            if let Some(p) = self.layer_progress.as_ref() {
+                p.send(ImportProgress::OstreeChunkCompleted(layer.layer.clone()))
+                    .await?;
+            }
         }
         if import.ostree_commit_layer.commit.is_none() {
+            if let Some(p) = self.layer_progress.as_ref() {
+                p.send(ImportProgress::OstreeChunkStarted(
+                    import.ostree_commit_layer.layer.clone(),
+                ))
+                .await?;
+            }
             let (blob, driver) = fetch_layer_decompress(
                 &mut self.proxy,
                 &self.proxy_img,
@@ -465,6 +496,12 @@ impl ImageImporter {
                 });
             let commit = super::unencapsulate::join_fetch(import_task, driver).await?;
             import.ostree_commit_layer.commit = Some(commit);
+            if let Some(p) = self.layer_progress.as_ref() {
+                p.send(ImportProgress::OstreeChunkCompleted(
+                    import.ostree_commit_layer.layer.clone(),
+                ))
+                .await?;
+            }
         };
         Ok(())
     }
@@ -511,6 +548,10 @@ impl ImageImporter {
                 tracing::debug!("Reusing fetched commit {}", c);
                 layer_commits.push(c.to_string());
             } else {
+                if let Some(p) = self.layer_progress.as_ref() {
+                    p.send(ImportProgress::DerivedLayerStarted(layer.layer.clone()))
+                        .await?;
+                }
                 let (blob, driver) = super::unencapsulate::fetch_layer_decompress(
                     &mut proxy,
                     &self.proxy_img,
@@ -533,6 +574,10 @@ impl ImageImporter {
                     let filtered = HashMap::from_iter(r.filtered.into_iter());
                     layer_filtered_content.insert(layer.digest().to_string(), filtered);
                 }
+                if let Some(p) = self.layer_progress.as_ref() {
+                    p.send(ImportProgress::DerivedLayerCompleted(layer.layer.clone()))
+                        .await?;
+                }
             }
         }
 
